from django.test import TestCase
from django.core.exceptions import ValidationError

from drip.models import Drip, QuerySetRule


class RulesTestCase(TestCase):
    def setUp(self):
        self.drip = Drip.objects.create(
            name='A Drip just for Rules',
            subject_template='Hello',
            body_html_template='KETTEHS ROCK!',
        )

    def test_valid_rule(self):
        rule = QuerySetRule(
            drip=self.drip,
            field_name='date_joined',
            lookup_type='lte',
            field_value='now-60 days',
        )
        rule.clean()

    def test_bad_field_name(self):
        rule = QuerySetRule(
            drip=self.drip,
            field_name='date__joined',
            lookup_type='lte',
            field_value='now-60 days',
        )
        self.assertRaises(ValidationError, rule.clean)

    def test_bad_field_value(self):
        rule = QuerySetRule(
            drip=self.drip,
            field_name='date_joined',
            lookup_type='lte',
            field_value='now-2 months',
        )
        self.assertRaises(ValidationError, rule.clean)



from datetime import timedelta

from django.test import TestCase
from django.test.client import RequestFactory
from django.urls import resolve, reverse
from django.utils import timezone

from drip.models import Drip, SentDrip, QuerySetRule
from drip.drips import DripBase
from drip.utils import get_user_model, unicode

from credits.models import Profile


class DripsTestCase(TestCase):

    def setUp(self):
        """
        Creates 20 users, half of which buy 25 credits a day,
        and the other half that does none.
        """
        self.User = get_user_model()

        start = timezone.now() - timedelta(hours=2)
        num_string = [
            'first',
            'second',
            'third',
            'fourth',
            'fifth',
            'sixth',
            'seventh',
            'eighth',
            'ninth',
            'tenth',
        ]

        for i, name in enumerate(num_string):
            user = self.User.objects.create(
                username='%s_25_credits_a_day' % name,
                email='%s@test.com' % name,
            )
            self.User.objects.filter(id=user.id).update(
                date_joined=start - timedelta(days=i),
            )

            profile = Profile.objects.get(user=user)
            profile.credits = i * 25
            profile.save()

        for i, name in enumerate(num_string):
            user = self.User.objects.create(
                username='%s_no_credits' % name,
                email='%s@test.com' % name,
            )
            self.User.objects.filter(id=user.id).update(
                date_joined=start - timedelta(days=i),
            )

    def test_users_exists(self):
        self.assertEqual(20, self.User.objects.all().count())

    def test_day_zero_users(self):
        start = timezone.now() - timedelta(days=1)
        end = timezone.now()
        self.assertEqual(
            2,
            self.User.objects.filter(
                date_joined__range=(start, end),
            ).count(),
        )

    def test_day_two_users_active(self):
        start = timezone.now() - timedelta(days=3)
        end = timezone.now() - timedelta(days=2)
        self.assertEqual(
            1,
            self.User.objects.filter(
                date_joined__range=(start, end),
                profile__credits__gt=0,
            ).count(),
        )

    def test_day_two_users_inactive(self):
        start = timezone.now() - timedelta(days=3)
        end = timezone.now() - timedelta(days=2)
        self.assertEqual(
            1,
            self.User.objects.filter(
                date_joined__range=(start, end),
                profile__credits=0,
            ).count(),
        )

    def test_day_seven_users_active(self):
        start = timezone.now() - timedelta(days=8)
        end = timezone.now() - timedelta(days=7)
        self.assertEqual(
            1, self.User.objects.filter(
                date_joined__range=(start, end),
                profile__credits__gt=0,
            ).count(),
        )

    def test_day_seven_users_inactive(self):
        start = timezone.now() - timedelta(days=8)
        end = timezone.now() - timedelta(days=7)
        self.assertEqual(
            1, self.User.objects.filter(
                date_joined__range=(start, end),
                profile__credits=0,
            ).count(),
        )

    def test_day_fourteen_users_active(self):
        start = timezone.now() - timedelta(days=15)
        end = timezone.now() - timedelta(days=14)
        self.assertEqual(
            0, self.User.objects.filter(
                date_joined__range=(start, end),
                profile__credits__gt=0,
            ).count(),
        )

    def test_day_fourteen_users_inactive(self):
        start = timezone.now() - timedelta(days=15)
        end = timezone.now() - timedelta(days=14)
        self.assertEqual(
            0, self.User.objects.filter(
                date_joined__range=(start, end),
                profile__credits=0,
            ).count(),
        )

    ########################
    #   RELATION SNAGGER   #
    ########################

    def test_get_simple_fields(self):
        from drip.utils import get_simple_fields

        simple_fields = get_simple_fields(self.User)
        self.assertTrue(
            bool([sf for sf in simple_fields if 'profile' in sf[0]]),
        )

    ##################
    #   TEST DRIPS   #
    ##################

    def test_backwards_drip_class(self):
        for drip in Drip.objects.all():
            self.assertTrue(issubclass(drip.drip.__class__, DripBase))

    def build_joined_date_drip(self, shift_one=7, shift_two=8):
        model_drip = Drip.objects.create(
            name='A Custom Week Ago',
            subject_template='HELLO {{ user.username }}',
            body_html_template='KETTEHS ROCK!',
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name='date_joined',
            lookup_type='lt',
            field_value='now-{0} days'.format(shift_one),
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name='date_joined',
            lookup_type='gte',
            field_value='now-{0} days'.format(shift_two),
        )
        return model_drip

    def test_custom_drip(self):
        """
        Test a simple
        """
        model_drip = self.build_joined_date_drip()
        drip = model_drip.drip

        # ensure we are starting from a blank slate
        # 2 people meet the criteria
        self.assertEqual(2, drip.get_queryset().count())
        drip.prune()
        # no one is pruned, never sent before
        self.assertEqual(2, drip.get_queryset().count())
        # confirm nothing sent before
        self.assertEqual(0, SentDrip.objects.count())

        # send the drip
        drip.send()
        self.assertEqual(2, SentDrip.objects.count())  # got sent

        for sent in SentDrip.objects.all():
            self.assertIn('HELLO', sent.subject)
            self.assertIn('KETTEHS ROCK', sent.body)

        # subsequent runs reflect previous activity
        drip = Drip.objects.get(id=model_drip.id).drip
        # 2 people meet the criteria
        self.assertEqual(2, drip.get_queryset().count())
        drip.prune()
        self.assertEqual(0, drip.get_queryset().count())  # everyone is pruned

    def test_custom_short_term_drip(self):
        model_drip = self.build_joined_date_drip(shift_one=3, shift_two=4)
        drip = model_drip.drip

        # ensure we are starting from a blank slate
        # 2 people meet the criteria
        self.assertEqual(2, drip.get_queryset().count())

    def test_custom_date_range_walk(self):
        model_drip = self.build_joined_date_drip()
        drip = model_drip.drip

        # vanilla (now-8, now-7), past (now-8-3, now-7-3),
        # future (now-8+1, now-7+1)
        for count, shifted_drip in zip(
            [0, 2, 2, 2, 2], drip.walk(into_past=3, into_future=2)
        ):
            self.assertEqual(
                count, shifted_drip.get_queryset().count()
            )

        # no reason to change after a send...
        drip.send()
        drip = Drip.objects.get(id=model_drip.id).drip

        # vanilla (now-8, now-7), past (now-8-3, now-7-3),
        # future (now-8+1, now-7+1)
        for count, shifted_drip in zip(
            [0, 2, 2, 2, 2], drip.walk(into_past=3, into_future=2)
        ):
            self.assertEqual(count, shifted_drip.get_queryset().count())

    def test_custom_drip_with_count(self):
        model_drip = self.build_joined_date_drip()
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name='profile__credits',
            lookup_type='gte',
            field_value='5',
        )
        drip = model_drip.drip

        # 1 person meet the criteria
        self.assertEqual(1, drip.get_queryset().count())

        for count, shifted_drip in zip(
            [0, 1, 1, 1, 1], drip.walk(into_past=3, into_future=2)
        ):
            self.assertEqual(count, shifted_drip.get_queryset().count())

    def test_exclude_and_include(self):
        model_drip = Drip.objects.create(
            name='A Custom Week Ago',
            subject_template='HELLO {{ user.username }}',
            body_html_template='KETTEHS ROCK!',
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name='profile__credits',
            lookup_type='gte',
            field_value='1',
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name='profile__credits',
            method_type='exclude',
            lookup_type='exact',
            field_value=100,
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name='profile__credits',
            method_type='exclude',
            lookup_type='exact',
            field_value=125,
        )
        # 7 people meet the criteria
        self.assertEqual(7, model_drip.drip.get_queryset().count())

    def test_custom_drip_static_datetime(self):
        model_drip = self.build_joined_date_drip()
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name='date_joined',
            lookup_type='lte',
            field_value=(
                timezone.now() - timedelta(days=8)
            ).strftime('%Y-%m-%d %H:%M:%S'),
        )
        drip = model_drip.drip

        for count, shifted_drip in zip(
            [0, 2, 2, 0, 0], drip.walk(into_past=3, into_future=2)
        ):
            self.assertEqual(count, shifted_drip.get_queryset().count())

    def test_custom_drip_static_now_datetime(self):
        model_drip = Drip.objects.create(
            name='A Custom Week Ago',
            subject_template='HELLO {{ user.username }}',
            body_html_template='KETTEHS ROCK!',
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name='date_joined',
            lookup_type='gte',
            field_value=(
                timezone.now() - timedelta(days=1)
            ).strftime('%Y-%m-%d 00:00:00'),
        )
        drip = model_drip.drip

        # catches "today and yesterday" users
        for count, shifted_drip in zip(
            [4, 4, 4, 4, 4], drip.walk(into_past=3, into_future=3)
        ):
            self.assertEqual(count, shifted_drip.get_queryset().count())

    def test_admin_timeline_prunes_user_output(self):
        """
        multiple users in timeline is confusing.
        """
        admin = self.User.objects.create(
            username='admin', email='admin@example.com')
        admin.is_staff = True
        admin.is_superuser = True
        admin.save()

        # create a drip campaign that will surely give us duplicates.
        model_drip = Drip.objects.create(
            name='A Custom Week Ago',
            subject_template='HELLO {{ user.username }}',
            body_html_template='KETTEHS ROCK!'
        )
        QuerySetRule.objects.create(
            drip=model_drip,
            field_name='date_joined',
            lookup_type='gte',
            field_value=(
                timezone.now() - timedelta(days=1)
            ).strftime('%Y-%m-%d 00:00:00'),
        )

        # then get it's admin view.
        rf = RequestFactory()
        timeline_url = reverse(
            'admin:drip_timeline',
            kwargs={
                'drip_id': model_drip.id,
                'into_past': 3,
                'into_future': 3,
            }
        )

        request = rf.get(timeline_url)
        request.user = admin

        match = resolve(timeline_url)

        response = match.func(request, *match.args, **match.kwargs)

        # check that our admin (not excluded from test) is shown once.
        self.assertEqual(unicode(response.content).count(admin.email), 1)

    ##################
    #   TEST M2M     #
    ##################

    def test_annotated_field_name_property_no_count(self):
        model_drip = Drip.objects.create(
            name='A Custom Week Ago',
            subject_template='HELLO {{ user.username }}',
            body_html_template='KETTEHS ROCK!',
        )

        qsr = QuerySetRule.objects.create(
            drip=model_drip,
            field_name='date_joined',
            lookup_type='exact',
            field_value=2,
        )
        self.assertEqual(qsr.annotated_field_name, 'date_joined')

    def test_annotated_field_name_property_with_count(self):

        model_drip = Drip.objects.create(
            name='A Custom Week Ago',
            subject_template='HELLO {{ user.username }}',
            body_html_template='KETTEHS ROCK!',
        )

        qsr = QuerySetRule.objects.create(
            drip=model_drip,
            field_name='userprofile__user__groups__count',
            lookup_type='exact',
            field_value=2,
        )

        self.assertEqual(
            qsr.annotated_field_name,
            'num_userprofile_user_groups',
        )

    def test_apply_annotations_no_count(self):

        model_drip = Drip.objects.create(
            name='A Custom Week Ago',
            subject_template='HELLO {{ user.username }}',
            body_html_template='KETTEHS ROCK!',
        )

        qsr = QuerySetRule.objects.create(
            drip=model_drip,
            field_name='date_joined',
            lookup_type='exact',
            field_value=2,
        )

        qs = qsr.apply_any_annotation(None)

        self.assertEqual(qs, None)

    def test_apply_annotations_with_count(self):

        model_drip = Drip.objects.create(
            name='A Custom Week Ago',
            subject_template='HELLO {{ user.username }}',
            body_html_template='KETTEHS ROCK!',
        )

        qsr = QuerySetRule.objects.create(
            drip=model_drip,
            field_name='profile__user__groups__count',
            lookup_type='exact',
            field_value=2,
        )

        qs = qsr.apply_any_annotation(model_drip.drip.get_queryset())
        self.assertEqual(
            list(qs.query.annotation_select.keys()),
            ['num_profile_user_groups'],
        )

    def test_apply_multiple_rules_with_aggregation(self):

        model_drip = Drip.objects.create(
            name='A Custom Week Ago',
            subject_template='HELLO {{ user.username }}',
            body_html_template='KETTEHS ROCK!',
        )

        qsr = QuerySetRule.objects.create(
            drip=model_drip,
            field_name='profile__user__groups__count',
            lookup_type='exact',
            field_value='0',
        )

        QuerySetRule.objects.create(
            drip=model_drip,
            field_name='date_joined',
            lookup_type='gte',
            field_value=(
                timezone.now() - timedelta(days=1)
            ).strftime('%Y-%m-%d 00:00:00'),
        )

        qsr.clean()
        qs = model_drip.drip.apply_queryset_rules(
            model_drip.drip.get_queryset()
        )

        self.assertEqual(qs.count(), 4)


class UrlsTestCase(TestCase):
    def test_drip_timeline_url(self):
        timeline_url = reverse(
            'admin:drip_timeline',
            kwargs={
                'drip_id': 1,
                'into_past': 2,
                'into_future': 3,
            }
        )

        self.assertEqual(timeline_url, '/admin/drip/drip/1/timeline/2/3/')

    def test_view_drip_email_url_url(self):
        view_drip_email_url = reverse(
            'admin:view_drip_email',
            kwargs={
                'drip_id': 1,
                'into_past': 2,
                'into_future': 3,
                'user_id': 4,
            }
        )

        self.assertEqual(
            view_drip_email_url,
            '/admin/drip/drip/1/timeline/2/3/4/',
        )



import re
import datetime

STRFDATETIME = re.compile('([dgGhHis])')


def STRFDATETIME_REPL(x):
    return '%%(%s)s' % x.group()


def process_regex(d):
    d = d.groupdict(0)
    if d['sign'] == '-':
        for k in 'hours', 'minutes', 'seconds':
            d[k] = '-' + d[k]
    d.pop('sign', None)
    return d


def get_flexible_regex(string):
    # This is the more flexible format
    d = re.match(
        r'^((?P<weeks>-?((\d*\.\d+)|\d+))\W*w((ee)?(k(s)?)?)(,)?\W*)?'
        r'((?P<days>-?((\d*\.\d+)|\d+))\W*d(ay(s)?)?(,)?\W*)?'
        r'((?P<hours>-?((\d*\.\d+)|\d+))\W*h(ou)?(r(s)?)?(,)?\W*)?'
        r'((?P<minutes>-?((\d*\.\d+)|\d+))\W*m(in(ute)?(s)?)?(,)?\W*)?'
        r'((?P<seconds>-?((\d*\.\d+)|\d+))\W*s(ec(ond)?(s)?)?)?\W*$',
        string,
    )
    if not d:
        raise TypeError("'%s' is not a valid time interval" % string)
    d = d.groupdict(0)
    return d


def process_string(string):
    # This is the format we get from sometimes Postgres, sqlite,
    # and from serialization
    d = re.match(
        r'^((?P<days>[-+]?\d+) days?,? )?(?P<sign>[-+]?)(?P<hours>\d+):'
        r'(?P<minutes>\d+)(:(?P<seconds>\d+(\.\d+)?))?$',
        str(string),
    )
    if d:
        d = process_regex(d)
    else:
        d = get_flexible_regex(string)
    return datetime.timedelta(**dict(((k, float(v)) for k, v in d.items())))


def parse(string):
    """
    Parse a string into a timedelta object.
    """
    string = string.strip()

    if string == "":
        raise TypeError("'%s' is not a valid time interval" % string)
    return process_string(string)



# Generated by Django 3.0.6 on 2020-05-29 20:38

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Drip',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('lastchanged', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(help_text='A unique name for this drip.', max_length=255, unique=True, verbose_name='Drip Name')),
                ('enabled', models.BooleanField(default=False)),
                ('from_email', models.EmailField(blank=True, help_text='Set a custom from email.', max_length=254, null=True)),
                ('from_email_name', models.CharField(blank=True, help_text='Set a name for a custom from email.', max_length=150, null=True)),
                ('subject_template', models.TextField(blank=True, null=True)),
                ('body_html_template', models.TextField(blank=True, help_text='You will have settings and user in the context.', null=True)),
                ('message_class', models.CharField(blank=True, default='default', max_length=120)),
            ],
        ),
        migrations.CreateModel(
            name='SentDrip',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('subject', models.TextField()),
                ('body', models.TextField()),
                ('from_email', models.EmailField(default=None, max_length=254, null=True)),
                ('from_email_name', models.CharField(default=None, max_length=150, null=True)),
                ('drip', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sent_drips', to='drip.Drip')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sent_drips', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='QuerySetRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now_add=True)),
                ('lastchanged', models.DateTimeField(auto_now=True)),
                ('method_type', models.CharField(choices=[('filter', 'Filter'), ('exclude', 'Exclude')], default='filter', max_length=12)),
                ('field_name', models.CharField(max_length=128, verbose_name='Field name of User')),
                ('lookup_type', models.CharField(choices=[('exact', 'exactly'), ('iexact', 'exactly (case insensitive)'), ('contains', 'contains'), ('icontains', 'contains (case insensitive)'), ('regex', 'regex'), ('iregex', 'contains (case insensitive)'), ('gt', 'greater than'), ('gte', 'greater than or equal to'), ('lt', 'less than'), ('lte', 'less than or equal to'), ('startswith', 'starts with'), ('endswith', 'starts with'), ('istartswith', 'ends with (case insensitive)'), ('iendswith', 'ends with (case insensitive)')], default='exact', max_length=12)),
                ('field_value', models.CharField(help_text='Can be anything from a number, to a string. Or, do `now-7 days` or `today+3 days` for fancy timedelta.', max_length=255)),
                ('drip', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='queryset_rules', to='drip.Drip')),
            ],
        ),
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
from os import path
import re
import os
import sys


current_dir = path.abspath(path.dirname('__file__'))
with open(path.join(current_dir, 'README.md'), encoding='utf-8') as file:
    long_description = file.read()


name = 'django-drip-campaigns'
package = 'drip'
description = 'Drip campaigns for Django.'
url = 'https://github.com/rootstrap/django-drip-campaigns'
author = 'Kalil de Lima'
author_email = 'kalil@rootstrap.com'
license = 'MIT'
install_requires = ['Django>=2.2', 'django-timedeltafield']
keywords = 'django drip email user query'


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search(
        "^__version__ = ['\"]([^'\"]+)['\"]", init_py, re.MULTILINE
    ).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}


if sys.argv[-1] == 'publish':
    os.system("python setup.py sdist upload")
    args = {'version': get_version(package)}
    print("You probably want to also tag the version now:")
    print("  git tag -a v%(version)s -m 'version v%(version)s'" % args)
    print("  git push --tags")
    sys.exit()


setup(
    name=name,
    version=get_version(package),
    url=url,
    license=license,
    description=description,
    author=author,
    author_email=author_email,
    keywords=keywords,
    packages=get_packages(package),
    package_data=get_package_data(package),
    install_requires=install_requires,
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



import sys

from django.db import models
from django.db.models import ForeignKey, OneToOneField, ManyToManyField
from django.db.models.fields.related import ForeignObjectRel as RelatedObject

# taking a nod from python-requests and skipping six
_ver = sys.version_info
is_py2 = (_ver[0] == 2)
is_py3 = (_ver[0] == 3)
basestring = None
unicode = None

if is_py2:
    basestring = basestring
    unicode = unicode
elif is_py3:
    basestring = (str, bytes)
    unicode = str


def check_redundant(model_stack, stack_limit):
    stop_recursion = False
    if len(model_stack) > stack_limit:
        # rudimentary CustomUser->User->CustomUser->User detection, or
        # stack depth shouldn't exceed x, or
        # we've hit a point where we are repeating models
        if (
            (model_stack[-3] == model_stack[-1]) or
            (len(model_stack) > 5) or
            (len(set(model_stack)) != len(model_stack))
        ):
            stop_recursion = True
    return stop_recursion


def get_field_name(field, RelatedObject):
    field_name = field.name

    if isinstance(field, RelatedObject):
        field_name = field.field.related_query_name()
    return field_name


def get_full_field(parent_field, field_name):
    if parent_field:
        full_field = "__".join([parent_field, field_name])
    else:
        full_field = field_name
    return full_field


def get_rel_model(field, RelatedObject):
    if isinstance(field, RelatedObject):
        RelModel = field.model
        # field_names.extend(get_fields(RelModel, full_field, True))
    else:
        RelModel = field.related_model
    return RelModel


def is_valid_instance(field):
    return (
        isinstance(field, ForeignKey) or
        isinstance(field, OneToOneField) or
        isinstance(field, RelatedObject) or
        isinstance(field, ManyToManyField)
    )


def get_out_fields(Model, parent_field, model_stack, excludes, fields):
    out_fields = []
    for field in fields:

        field_name = get_field_name(field, RelatedObject)

        full_field = get_full_field(parent_field, field_name)

        if len([True for exclude in excludes if (exclude in full_field)]):
            continue

        # add to the list
        out_fields.append([full_field, field_name, Model, field.__class__])

        if is_valid_instance(field):

            RelModel = get_rel_model(field, RelatedObject)

            out_fields.extend(
                get_fields(RelModel, full_field, list(model_stack)),
            )

    return out_fields


def get_fields(
    Model,
    parent_field="",
    model_stack=[],
    stack_limit=2,
    excludes=['permissions', 'comment', 'content_type']
):
    """
    Given a Model, return a list of lists of strings with important stuff:
    ...
    ['test_user__user__customuser', 'customuser', 'User', 'RelatedObject']
    ['test_user__unique_id', 'unique_id', 'TestUser', 'CharField']
    ['test_user__confirmed', 'confirmed', 'TestUser', 'BooleanField']
    ...

     """

    # github.com/omab/python-social-auth/commit/d8637cec02422374e4102231488481170dc51057
    if isinstance(Model, basestring):
        app_label, model_name = Model.split('.')
        Model = models.get_model(app_label, model_name)

    fields = Model._meta.fields + \
        Model._meta.many_to_many + \
        Model._meta.get_fields()
    model_stack.append(Model)

    # do a variety of checks to ensure recursion isnt being redundant

    stop_recursion = check_redundant(model_stack, stack_limit)

    if stop_recursion:
        return []  # give empty list for "extend"

    return get_out_fields(Model, parent_field, model_stack, excludes, fields)


def give_model_field(full_field, Model):
    """
    Given a field_name and Model:

    "test_user__unique_id", <AchievedGoal>

    Returns "test_user__unique_id", "id", <Model>, <ModelField>
    """
    field_data = get_fields(Model, '', [])

    for full_key, name, _Model, _ModelField in field_data:
        if full_key == full_field:
            return full_key, name, _Model, _ModelField

    raise Exception('Field key `{0}` not found on `{1}`.'.format(
        full_field, Model.__name__),
    )


def get_simple_fields(Model, **kwargs):
    ret_list = []
    for f in get_fields(Model, **kwargs):
        if '__' in f[0]:
            # Add __user__ to the fields in related models
            parts = f[0].split('__')
            f[0] = parts[0] + '__user__' + parts[1]
        if f[0] not in [x[0] for x in ret_list]:
            # Add field if not already in list
            ret_list.append([f[0], f[3].__name__])
    return ret_list


def get_user_model():
    # handle 1.7 and back
    try:
        from django.contrib.auth import get_user_model as django_get_user_model
        User = django_get_user_model()
    except ImportError:
        from django.contrib.auth.models import User
    return User
